
#include <math.h>
#include <avr/eeprom.h>

#include "xavr/xavr.h"
#include "xavr/mmc.h"

ISR(TIMER1_OVF_vect)
{
  XAVR::Ticker::Pass(1);
}

template <class DATA, class BCK, class WS>
class I2S
{
public:
  static void Init ()
  {
    XAVR::Pins<DATA, BCK, WS>::Init();
  }

  static inline void send_bit(uint16_t n, uint8_t pos) _XAVR_INLINE
  {
    if ( n & (1 << pos) ) DATA::Set(); else DATA::Clear();
    BCK::Set(); BCK::Clear();
  }

  static void Send16 (uint16_t n)
  {
    send_bit (n, 15);
    send_bit (n, 14);
    send_bit (n, 13);
    send_bit (n, 12);
    send_bit (n, 11);
    send_bit (n, 10);
    send_bit (n, 9);
    send_bit (n, 8);
    send_bit (n, 7);
    send_bit (n, 6);
    send_bit (n, 5);
    send_bit (n, 4);
    send_bit (n, 3);
    send_bit (n, 2);
    send_bit (n, 1);
    send_bit (n, 0);

    // switch to the other channel but really don't send anything
    // to save time, tda1543 can take it
    WS::Clear();
    BCK::Set();
    BCK::Clear();
    WS::Set();
  }
};

static volatile enum { PLAYING, RECORDING } estatus = PLAYING;

typedef I2S<XAVR::PortD::Out0, XAVR::PortD::Out2, XAVR::PortD::Out1> DAC;
typedef XAVR::PortC::In5 BUTTON;
static EEMEM uint32_t eeprom_count;
static uint32_t count = 50000;

static bool buttonIs (bool onoff)
{
  bool bs = BUTTON::IsOn();
  if ( onoff && !bs || !onoff && bs )
      return false;
  for ( uint32_t j = 0; j < 300; j++ ) // wait for possible bounce to settle
      __asm__ __volatile__ ("nop");
  if ( onoff && !bs || !onoff && bs )
      return false;
  return true;
}

static void beep()
{
  for ( uint8_t i = 0; i < 8; i++ )
    {
      #define RR 512
      #define BB 34256
      for ( uint16_t n = BB - RR/2; n <= BB + RR/2; n += 32 )
          DAC::Send16(n);
      for ( uint16_t n = BB + RR/2; n >= BB - RR/2; n -= 32 )
          DAC::Send16(n);
    }
}

void error(uint32_t n, uint32_t n2 = 0)
{
  cli();
  while (1)
    {
      for ( uint32_t i = 0; i < n * 2; i++ )
        {
          for ( uint32_t j = 0; j < 150000; j++ )
              __asm__ __volatile__ ("nop");
          if ( i & 1 ) beep();
        }

      for ( uint32_t j = 0; j < 400000; j++ )
          __asm__ __volatile__ ("nop");

      if ( n2 )
        {
          for ( uint32_t i = 0; i < n2 * 2; i++ )
            {
              for ( uint32_t j = 0; j < 150000; j++ )
                  __asm__ __volatile__ ("nop");
              if ( i & 1 ) beep();
            }

          for ( uint32_t j = 0; j < 400000; j++ )
              __asm__ __volatile__ ("nop");
        }
    }
}

class Buffer
{
public:
  uint8_t * volatile wp;
  uint8_t * volatile rp;

#define BUFLEN 320

  uint8_t BUF[BUFLEN];
  uint8_t *END;
  uint32_t cnt;

  Buffer()
  {
    END = BUF + BUFLEN;
    Reset();
  }

  void Reset ()
  {
    wp = ((uint8_t*)BUF) + 1;
    rp = BUF;
    *rp = 127;
    cnt = count;
  }

  // there is one big problem with this code - the wp and rp
  // are read and written from either main loop or from
  // an interrrupt handler - so bad things can happen - like
  // relooping over a buffer of data
  inline void Write (uint8_t v, bool block = true) _XAVR_INLINE
  {
    if ( block )
        while ( wp == rp );
    else if ( wp == rp )
        return;
    *wp = v;
    wp++;
    if ( wp == (uint8_t*)END )
        wp = (uint8_t*)BUF;
  }

  inline uint8_t Read (bool block = true) _XAVR_INLINE
  {
    uint8_t ret = *rp;
    if ( block )
        while ( rp + 1 == wp || rp + 1 == END && wp == BUF );
    if ( !( rp + 1 == wp || rp + 1 == END && wp == BUF ) )
      {
        rp++;
        if ( rp == END )
            rp = BUF;
      }
    if ( estatus == RECORDING )
        count++;
    else if ( cnt != 0 )
        cnt--;
    return ret;
  }

  inline bool Closed() _XAVR_INLINE
  {
    return estatus == PLAYING && (buttonIs(true) || cnt == 0) || estatus == RECORDING && buttonIs(false);
  }
};

static Buffer buffer;

#define BASE 34256
#define RANGE 512

static bool roll = false;
static uint8_t last_sample;
ISR(TIMER0_OVF_vect)
{
  if ( !roll )
    {
      roll = true;
      return;
    }
  roll = false;

  if ( estatus == PLAYING )
    {
      uint8_t v = (uint8_t)255 - buffer.Read(false);
      DAC::Send16 (BASE - RANGE/2 + (uint16_t)v * (RANGE/256));
    }
  else
    {
      buffer.Write (last_sample, false);
    }
}

static uint16_t calibrate_mid = 0;
static uint16_t calibrate_cnt = 0;
static uint32_t calibrate_sum = 0;

static void set_calibrate ()
{
  calibrate_cnt = 0;
  calibrate_sum = 0;
  calibrate_mid = 0;
}

ISR(ADC_vect)
{
  uint16_t v = ADC;

  if ( calibrate_cnt < 1000 )
    {
      calibrate_cnt++;
      if ( calibrate_cnt < 0xff )
          return;
      calibrate_sum += v;
      return;
    }

  if ( calibrate_mid == 0 )
      calibrate_mid = calibrate_sum / (1000 - 0xff);

  if ( estatus != RECORDING )
      return;

  // effectively 7 bits?
  int16_t r = ((int16_t)v - (int16_t)calibrate_mid) * 2;
  if ( r < -127 )
      v = 0;
  else if ( r > 128 )
      v = 0xff;
  else
      v = r + 127;

  last_sample = v;
}

int main ()
{
  for ( uint32_t j = 0; j < 10000; j++ ) // give time for caps to charge
      __asm__ __volatile__ ("nop");

  DAC::Init();
  BUTTON::Init();

  ADMUX = _BV(MUX0);
  DIDR0 = _BV(ADC1D);
  ADCSRA = _BV(ADEN) | _BV(ADPS2) | _BV(ADPS0) |  _BV(ADIE) | _BV(ADATE);
  ADCSRB = 0;

  // sample/play timer
  TCCR0A = 0;
  TCCR0B = 0x01;
  TIMSK0 |= 0x01;

  // ticker
  TCCR1A = 0;
  TCCR1B = 0x01;
  TIMSK1 |= 0x01;

  eeprom_read_block (&count, &eeprom_count, 4);

  sei();

  // initialize the MMC
  XAVR::MMC<> mmc;
  uint8_t res = mmc.Init();
  if ( res != mmc.initOk )
      error (res, 1);

  while (1)
    {
      if ( BUTTON::IsOn() )
        {
          cli();
          set_calibrate();
          estatus = RECORDING;
          buffer.Reset();
          ADCSRA |= _BV(ADEN); // enable AD converter
          ADCSRA |= _BV(ADSC); // start sampling
          count = 0;
          sei();
          if ( !mmc.WriteMMC<> (buffer, 0) )
              error (2);
          cli();
          ADCSRA &= ~_BV(ADEN); // disable AD converter
          eeprom_write_block (&count, &eeprom_count, 4);
          sei();
        }
      else
        {
          cli();
          estatus = PLAYING;
          buffer.Reset();
          sei();
          if ( !mmc.ReadMMC<> (buffer, 0) )
              error (3);
        }
    }

  return 0;
}

