
#ifndef __mmc_h__
#define __mmc_h__

namespace XAVR
{

class GenericMMCInterface
{
public:
  static inline bool InitSPI (uint8_t cno) _XAVR_INLINE
  {
    SPI::InitMaster(SPI::DIV_2);
    return true;
  }

  static inline bool InitSS (uint8_t cno) _XAVR_INLINE
  {
    SPI::OutSS::Init();
    return true;
  }

  static inline void SelectSS (uint8_t cno, bool sel) _XAVR_INLINE
  {
    if ( sel )
        SPI::OutSS::Clear();
    else
        SPI::OutSS::Set();
  }

  static inline void Send (uint8_t cno, uint8_t byte) _XAVR_INLINE
  {
    SPI::Send (byte);
  }

  static inline uint8_t Recv (uint8_t cno) _XAVR_INLINE
  {
    return SPI::SendRecv (0xff);
  }
};

template <class Interface = GenericMMCInterface>
class MMC
{
public:
  MMC(uint8_t cno = 0);

  struct AutoDeselect
  {
    inline AutoDeselect (MMC *mmc) _XAVR_INLINE : mmc(mmc) { }
    inline ~AutoDeselect () _XAVR_INLINE { mmc->SelectSS(false); }
    MMC *mmc;
  };

  inline void SelectSS (bool sel) _XAVR_INLINE
  {
    Interface::SelectSS (cno, sel);
  };

  inline void Send (uint8_t byte) _XAVR_INLINE
  {
    Interface::Send (cno, byte);
  }

  inline uint8_t Recv () _XAVR_INLINE
  {
    return Interface::Recv (cno);
  };

  enum eInitResponse
  {
    initOk = 0,
    initSPIInitFailed,
    initCardNotOnline,
    initCardNotResponding,
    initCardTimeout,
    initCardInvalidVoltage,
    initCardInvalidResponse,
    initUnsupportedBlockSize,
  };

  enum
  {
    GO_IDLE_STATE         = 0x40 + 0,   // 0,
    SEND_OP_COND_MMC      = 0x40 + 1,   // 1,
    SEND_IF_COND          = 0x40 + 8,   // 8
    STOP_TRANSMISSION     = 0x40 + 12,  // 12
    SET_BLOCKLEN          = 0x40 + 16,  // 16
    READ_SINGLE_BLOCK     = 0x40 + 17,  // 17
    READ_MULTIPLE_BLOCK   = 0x40 + 18,  // 18
    SET_WR_BLK_ERASE_COUNT= 0xC0 + 23,  // 23 SDC
    WRITE_SINGLE_BLOCK    = 0x40 + 24,  // 24
    WRITE_MULTIPLE_BLOCK  = 0x40 + 25,  // 25
    SEND_OP_COND_SDC      = 0xC0 + 41,  // 41, SDC
    APP_CMD               = 0x40 + 55,  // 55
    READ_OCR              = 0x40 + 58,  // 58
  };

  eInitResponse Init();

  bool Read (uint32_t sector, uint8_t *buff, uint8_t count);
  bool Write (uint32_t sector, const uint8_t *buff, uint8_t count);

  operator bool () const { return online; }

  uint8_t SendCommand (uint8_t cmd, uint32_t arg);

  bool WaitReady();

  /**
    * Write data from the stream into the MMC. The Stream class must have following methods:
    *   uint8_t Read();
    *   bool Closed();
    * Writing is done until Empty() gives false.
    */
  template <class Stream>
  bool WriteMMC (Stream& src, uint32_t sid)
  {
    if ( !online )
        return false;
    while ( !src.Closed() )
      {
        uint32_t sector = sid;
        if ( !(type & 8) )
            sector *= 512;
        if ( SendCommand (WRITE_SINGLE_BLOCK, sector) != 0  )
            return false;
        if ( !WaitReady() )
            return false;
        Send (0xFE);
        for ( uint8_t i = 0; i < 128; i++ )
          {
            Send (src.Read()); Send (src.Read());
            Send (src.Read()); Send (src.Read());
          }
        Send (0xff);
        Send (0xff);
        if ( (Recv() & 0x1F) != 0x05 )
            return false;
        sid++;
      }
    return true;
  }

  /**
    * Read MMC data and write into a stream. The Stream class must have following methods:
    *   void Write(uint8_t b);
    *   bool Closed();
    * Writing is done until Closed() gives false.
    */
  template <class Stream>
  bool ReadMMC (Stream& dst, uint32_t sid)
  {
    if ( !online )
        return false;

    if ( SendCommand (READ_MULTIPLE_BLOCK, sid) != 0  )
        return false;

    while ( !dst.Closed() )
      {
        XAVR::Ticker::Reset();
        while (1)
          {
            if ( Recv() == 0xFE )
                break;
            if ( XAVR::Ticker::Out() )
                return false;
          }

        for ( uint8_t i = 0; i < 128; i++ )
          {
            dst.Write (Recv());
            dst.Write (Recv());
            dst.Write (Recv());
            dst.Write (Recv());
          }

        Recv();
        Recv();
      }

    SendCommand (STOP_TRANSMISSION, 0);

    return true;
  }

private:
  bool ReadBlock (uint8_t *ptr, uint16_t cnt);
  bool WriteBlock (uint8_t token, const uint8_t *ptr, uint16_t cnt);

  bool online;
  uint8_t type;
  uint8_t cno;
};

template <class Interface>
MMC<Interface>::MMC (uint8_t cno) : online(false), cno(cno)
{
}

template <class Interface>
typename MMC<Interface>::eInitResponse MMC<Interface>::Init()
{
  // initialize the SPI interface and the SS (Slave-Select) pin
  if ( !Interface::InitSS(cno) || !Interface::InitSPI(cno) )
      return initSPIInitFailed;

  AutoDeselect adsl(this);

  // waiting 80 spi cycles for things to go online
  SelectSS(false);
  for ( uint8_t i = 0; i < 80; i++ )
      Recv();

  if ( SendCommand (GO_IDLE_STATE, 0) != 1 )
      return initCardNotResponding;

  if ( SendCommand (SEND_IF_COND, 0x1AA) == 1 ) // SDHC
    {
      uint8_t ocr[4];
      ocr[0] = Recv(); ocr[1] = Recv();
      ocr[2] = Recv(); ocr[3] = Recv();
      // can work at our vdd range (2.7-3.6V)
      if ( !(ocr[2] == 0x01 && ocr[3] == 0xAA) )
          return initCardInvalidVoltage;

      // wait for leaving the idle state
      XAVR::Ticker::Reset();
      while (1)
        {
          if ( !SendCommand (SEND_OP_COND_SDC, 1UL << 30) )
              break;
          if ( XAVR::Ticker::Out() )
              return initCardTimeout;
        }

      if ( SendCommand (READ_OCR, 0) != 0 )
          return initCardInvalidResponse;

      ocr[0] = Recv(); ocr[1] = Recv();
      ocr[2] = Recv(); ocr[3] = Recv();

      type = (ocr[0] & 0x40) ? 12 : 4;
    }
  else
    {
      if ( SendCommand (SEND_OP_COND_SDC, 0) <= 1 )
        {
          // SDSC
          while ( SendCommand (SEND_OP_COND_SDC, 0) );
          type = 2;
        }
      else
        {
          // MMC
          while ( SendCommand (SEND_OP_COND_MMC, 0) );
          type = 1;
        }

      if ( SendCommand (SET_BLOCKLEN, 512) != 0 )
          return initUnsupportedBlockSize;
    }

  online = true;
  return initOk;
}

template <class Interface>
bool MMC<Interface>::Read (uint32_t sector, uint8_t *buff, uint8_t count)
{
  if ( !online || !count )
      return false;
  AutoDeselect adsl(this);
  if ( !(type & 8) ) sector *= 512;
  if ( count == 1 )
    {
      if ( SendCommand (READ_SINGLE_BLOCK, sector) != 0  )
          return false;
      if ( !ReadBlock (buff, 512) )
          return false;
    }
  else
    {
      if ( SendCommand (READ_MULTIPLE_BLOCK, sector) != 0  )
          return false;
      while ( count )
        {
          if ( !ReadBlock (buff, 512) )
              return false;
          buff += 512;
          count--;
        }
      SendCommand (STOP_TRANSMISSION, 0);
    }
  return true;
}

template <class Interface>
bool MMC<Interface>::Write (uint32_t sector, const uint8_t *buff, uint8_t count)
{
  if ( !online || !count )
      return false;
  AutoDeselect adsl(this);
  if ( !(type & 8) ) sector *= 512;
  if ( count == 1 )
    {
      if ( SendCommand (WRITE_SINGLE_BLOCK, sector) != 0  )
          return false;
      if ( !WriteBlock (0xFE, buff, 512) )
          return false;
    }
  else
    {
      if ( type & 6 )
          SendCommand (SET_WR_BLK_ERASE_COUNT, count);
      if ( SendCommand (WRITE_MULTIPLE_BLOCK, sector) != 0  )
          return false;
      while ( count )
        {
          if ( !WriteBlock (0xFC, buff, 512) )
              return false;
          buff += 512;
          count--;
        }
      WriteBlock (0xFD, 0, 0);
    }
  return true;
}

template <class Interface>
bool MMC<Interface>::WaitReady()
{
  Send(0xff);
  XAVR::Ticker::Reset();
  while(1)
    {
      if ( Recv() == 0xff )
          return true;
      if ( XAVR::Ticker::Out() )
          return false;
    }
}

template <class Interface>
uint8_t MMC<Interface>::SendCommand (uint8_t cmd, uint32_t arg)
{
  if ( cmd & 0x80 )
    {
      cmd &= 0x7f;
      uint8_t res = SendCommand (APP_CMD, 0);
      if ( res != 1 )
          return res;
    }

  SelectSS(false); // reset MMC selection just in case it was not done before
  Recv(); // give time for the card to reset
  SelectSS(true); // select the MMC card
  WaitReady(); // and wait for ready

  // send the command
  Send (cmd);
  Send ((uint8_t)(arg >> 24));
  Send ((uint8_t)(arg >> 16));
  Send ((uint8_t)(arg >> 8));
  Send ((uint8_t)(arg >> 0));

  // send the crc
  if ( cmd == GO_IDLE_STATE )
      Send (0x95);
  else if ( cmd == SEND_IF_COND )
      Send (0x87);
  else
      Send (0x01);

  // stop transmission, skip some sort of staff byte
  if ( cmd == STOP_TRANSMISSION )
      Recv ();

  // receive command response
  uint8_t res;
  for ( uint8_t i = 0; i < 10; i++ )
    {
      res = Recv ();
      if ( ! (res & 0x80) )
          break;
    }
  return res;
}

template <class Interface>
bool MMC<Interface>::ReadBlock (uint8_t *ptr, uint16_t cnt)
{
  uint8_t token;
  // receive data start indicator
  XAVR::Ticker::Reset();
  while (1)
    {
      token = Recv();
      if ( token == 0xFE )
          break;
      if ( XAVR::Ticker::Out() )
          return false;
    }
  // receive data
  while (cnt)
    {
      *ptr++ = Recv();
      *ptr++ = Recv();
      *ptr++ = Recv();
      *ptr++ = Recv();
      cnt -= 4;
    }
  // CRC dummy
  Recv();
  Recv();
  return true;
}

template <class Interface>
bool MMC<Interface>::WriteBlock (uint8_t token, const uint8_t *ptr, uint16_t cnt)
{
  if ( !WaitReady() )
      return false;
  // send token
  Send (token);
  // special empty token
  if ( token == 0xFD )
      return true;
  // send data
  while (cnt)
    {
      Send (*ptr++);
      Send (*ptr++);
      Send (*ptr++);
      Send (*ptr++);
      cnt -= 4;
    }
  // CRC dummy
  Send (0xff);
  Send (0xff);
  return (Recv() & 0x1F) == 0x05;
}

}; // namespace XAVR

#endif // !__mmc_h__

